"""Test user scraping."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from untappd_scraper.user import User
from untappd_scraper.user_lists import WebUserList
from utpd_models_web.constants import UNTAPPD_BEER_HISTORY_SIZE

# sourcery skip: dont-import-test-modules
if TYPE_CHECKING:  # pragma: no cover
    from tests.conftest import MockResponse


@pytest.fixture
def user(
    _mock_user_get: None,
    # _mock_user_beer_history_get: None,
    _mock_user_venue_history_get: None,
    _mock_user_lists_get: None,
) -> User:
    return User("test")


# ----- Tests -----


def test_user(user: User) -> None:
    result = user

    assert result
    assert result.name
    assert result.user_id
    assert result.user_name


@pytest.mark.httpbin
@pytest.mark.usefixtures("_mock_user_404")
def test_user_invalid() -> None:
    with pytest.raises(ValueError, match="Invalid"):
        User("123")  # ignored


def test_activity(user: User) -> None:
    result = user.activity()

    assert result
    assert result[0].checkin_id
    assert result[0].beer_id
    assert all("jpeg" in r.beer_label_url for r in result)
    assert all(r.brewery_url and "tba" not in r.brewery_url.casefold() for r in result)


@pytest.mark.webtest
def test_activity_mw1414() -> None:
    result = User("mw1414").activity()

    assert result


@pytest.fixture(scope="module")
def wardy_user() -> User:
    """Create a real, web-scraped user."""
    return User("mw1414")


@pytest.mark.webtest
def test_beer_history(wardy_user: User) -> None:
    result = wardy_user.beer_history()

    assert result
    assert len(result.results) == UNTAPPD_BEER_HISTORY_SIZE
    assert len({b.beer_id for b in result.results}) == UNTAPPD_BEER_HISTORY_SIZE
    assert result.results[0].beer_name
    assert result.results[0].brewery_name
    assert result.results[0].recent_checkin_id
    assert not result.found_all  # way more than one page of uniques
    assert "total_found=25," in repr(result)
    assert all("assets" in r.beer_label_url for r in result.results)
    assert all(r.brewery_url and r.brewery_url != "TBA" for r in result.results)
    assert all(r.beer_label_url for r in result.results)


@pytest.mark.webtest
def test_beer_brewery_history_alesmith(wardy_user: User) -> None:
    # As of Apr 2025, Wardy had 31 Alesmith beers
    result = wardy_user.brewery_history(brewery_id=2471, max_resorts=1)

    assert result
    assert result.total_found
    assert result.found_all
    assert all(r.beer_label_url for r in result.results)


@pytest.mark.webtest
def test_beer_brewery_history_buckettys(wardy_user: User) -> None:
    # As of Apr 2025, Wardy had 84 bucketty's beers
    # and an English IPA was elusive in re-sorts
    result = wardy_user.brewery_history(brewery_id=484738, max_resorts=99)

    assert result
    assert len(result.results) > UNTAPPD_BEER_HISTORY_SIZE
    assert len(result.results) == 84  # NOTE fragile!
    assert result.found_all
    assert all(r.beer_label_url for r in result.results)


# NOTE this is huge. Probably should skip of limit resorts
@pytest.mark.webtest
def test_beer_brewery_history_4pines(wardy_user: User) -> None:
    max_resorts = 3
    # As of Apr 2025, Wardy had 569 4 Pines beers!!
    result = wardy_user.brewery_history(brewery_id=4254, max_resorts=max_resorts)

    assert result
    assert result.total_expected == 569  # NOTE fragile!
    # expect more than just the pages we requested, as style sort will grab more
    assert result.total_found > UNTAPPD_BEER_HISTORY_SIZE * max_resorts
    assert not result.found_all
    assert all(r.beer_label_url for r in result.results)


def test_lists(user: User) -> None:
    lists = user.lists()
    assert len(lists) == 13

    result = lists[0]

    assert result.description
    assert result.num_items


def test_lists_detail(
    user: User,
    list_page_1_resp: MockResponse,
    list_page_2_resp: MockResponse,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    lists = user.lists()
    fridge = lists[1]  # should be fridge
    monkeypatch.setattr(
        "untappd_scraper.user_lists.list_page_all_sorts",
        lambda _: (list_page_1_resp, list_page_2_resp),
    )  # pyright: ignore[reportCallIssue]
    lists = user.lists_detail(fridge.name)
    assert len(lists) == 1

    result = lists[0]

    assert isinstance(result, WebUserList)
    assert len(result.beers) == result.num_items
    assert result.full_scrape


def test_venue_history(user: User) -> None:
    history = user.venue_history()
    assert len(history) == UNTAPPD_BEER_HISTORY_SIZE

    result = next(iter(history))

    assert result.venue_id
    assert result.first_checkin_id
    assert result.last_checkin_id



"""Untappd beers."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from dateutil.parser import parse as parse_date
from requests_html import Element
from utpd_models_web.beer import WebBeerDetails
from utpd_models_web.checkin import WebActivityBeer

from untappd_scraper.html_session import get
from untappd_scraper.logging_config import logger
from untappd_scraper.web import id_from_href, parse_abv, parsed_value, slug_from_href

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Iterator

    from requests_html import HTMLResponse


@dataclass
class Beer:
    """Untappd beer."""

    beer_id: int
    beer_name: str = field(init=False)

    _page: HTMLResponse = field(init=False, repr=False)
    _beer_details: WebBeerDetails = field(init=False, repr=False)

    def __post_init__(self) -> None:
        """Post-init method to load beer details."""
        self._page = get(url_of(self.beer_id))  # pyright: ignore[reportArgumentType, reportAttributeAccessIssue]
        if not self._page.ok:
            msg = f"Invalid beer ID {self.beer_id} ({self._page})"
            raise ValueError(msg)

        self._beer_details = beer_details(resp=self._page)
        self.beer_name = self._beer_details.beer_name

    def __getattr__(self, name: str) -> Any:  # noqa: ANN401
        """Return unknown attributes from beer details.

        Args:
            name (str): attribute to lookup

        Returns:
            Any: attribute value
        """
        return getattr(self._beer_details, name)


# ----- utils -----


def url_of(beer_id: int) -> str:
    """Return the URL for a beer's main page.

    Args:
        beer_id (int): beer ID

    Returns:
        str: url to load to get beer's main page
    """
    return f"https://untappd.com/beer/{beer_id}"


# ----- beer details processing -----


def beer_details(resp: HTMLResponse) -> WebBeerDetails:
    """Parse a user's main page into user details.

    Args:
        resp (HTMLResponse): beer's main page loaded

    Returns:
        WebBeerDetails: general beer details
    """
    content_el = resp.html.find(".main .content", first=True)
    description = "".join(  # pyright: ignore[reportCallIssue]
        content_el.find(".desc .beer-descrption-read-less", first=True).xpath("//div/text()")  # pyright: ignore[reportAttributeAccessIssue, reportArgumentType]
    ).strip()

    return WebBeerDetails(
        beer_id=id_from_href(content_el.find("a.check", first=True)),  # pyright: ignore[reportAttributeAccessIssue, reportArgumentType]
        beer_name=content_el.find(".name h1", first=True).text,  # pyright: ignore[reportAttributeAccessIssue]
        beer_description=description.strip(),
        beer_label_url=content_el.find(".label img", first=True).attrs["src"],  # pyright: ignore[reportAttributeAccessIssue]
        brewery_name=content_el.find(".name .brewery a", first=True).text,  # pyright: ignore[reportAttributeAccessIssue]
        brewery_url=(content_el.find(".name .brewery a", first=True))  # pyright: ignore[reportAttributeAccessIssue]
        .attrs["href"]  # pyright: ignore[reportAttributeAccessIssue]
        .removesuffix("/"),
        style=content_el.find(".name p.style", first=True).text,  # pyright: ignore[reportAttributeAccessIssue]
        abv=parse_abv(content_el.find(".details p.abv", first=True).text),  # pyright: ignore[reportAttributeAccessIssue]
        url=resp.url,
        global_rating=content_el.find(".details [data-rating]", first=True).attrs[  # pyright: ignore[reportAttributeAccessIssue]
            "data-rating"
        ],
        num_ratings=parsed_value(  # pyright: ignore[reportArgumentType]
            "{:d} Rat",
            content_el.find(".details p.raters", first=True).text,  # pyright: ignore[reportAttributeAccessIssue]
        ),
    )


def checkin_activity(resp: HTMLResponse) -> Iterator[WebActivityBeer]:
    """Parse all available recent checkins for a user or in a venue.

    Args:
        resp (HTMLResponse): user's main page or venue's activity page

    Returns:
        Iterator[WebActivityBeer]: user's visible recent checkins
    """
    return (checkin_details(checkin) for checkin in resp.html.find(".activity .item"))  # pyright: ignore[reportGeneralTypeIssues]


def checkin_details(checkin_item: Element) -> WebActivityBeer:
    """Extract beer details from a checkin.

    Args:
        checkin_item (Element): single checkin

    Returns:
        WebActivityBeer: Interesting details for a beer
    """
    user_el, beer_el, brewery_el, location_el = extract_checkin_elements(checkin_item)

    checkin_time_el = checkin_item.find(".bottom .time", first=True)
    checkin_time = checkin_time_el.attrs.get("data-gregtime", checkin_time_el.text)  # pyright: ignore[reportAttributeAccessIssue]
    checkin_time = parse_date(checkin_time)
    assert checkin_time.tzinfo
    assert checkin_time.tzinfo.utcoffset(checkin_time) is not None, (
        f"Naive datetime from {checkin_time_el.html=}"  # pyright: ignore[reportAttributeAccessIssue]
    )

    if (img := checkin_item.find("a.label img", first=True)) and isinstance(img, Element):
        label = img.attrs["src"]
    else:
        label = ""  # pragma: no cover

    purchased_at = checkin_item.find(".purchased a", first=True)
    try:
        comment = checkin_item.find("p.comment-text", first=True).text  # pyright: ignore[reportAttributeAccessIssue]
    except AttributeError:
        comment = None
    serving = checkin_item.find(".serving", first=True)

    data_rating_element = checkin_item.find("[data-rating]", first=True)
    if data_rating_element:
        data_rating: float | None = float(data_rating_element.attrs["data-rating"])  # pyright: ignore[reportAttributeAccessIssue]
    else:
        data_rating = None  # pragma: no cover

    try:
        friends: list[str] | None = [
            slug_from_href(href)
            for href in checkin_item.find(".tagged-friends a")  # pyright: ignore[reportGeneralTypeIssues]
        ]
    except AttributeError:  # pragma: no cover
        friends = None

    if brewery_link := brewery_el.find("a", first=True):
        brewery_url: str = brewery_link.attrs["href"].strip("/")  # pyright: ignore[reportAttributeAccessIssue]

        last_bit = brewery_url.rsplit("/", 1)[-1]
        brewery_id = int(last_bit) if last_bit.isdigit() else None
    else:
        logger.error("No brewery link found in %s", brewery_el.html)
        msg = f"No brewery link found for beer {beer_el.text}"
        raise ValueError(msg)  # pragma: no cover

    try:
        beer_id = id_from_href(beer_el)
    except ValueError:
        logger.exception("Failed to parse beer ID from %s", beer_el.html)
        beer_id = 0

    return WebActivityBeer(
        checkin_id=int(checkin_item.attrs["data-checkin-id"]),
        checkin=checkin_time,
        user_name=slug_from_href(user_el),
        beer_id=beer_id,
        beer_label_url=label,
        beer_name=beer_el.text,
        brewery_name=brewery_el.text,
        brewery_id=brewery_id,
        brewery_url=brewery_url,
        location=location_el.text if location_el else None,
        location_id=id_from_href(location_el) if location_el else None,
        purchased_at=purchased_at.text if purchased_at else None,  # pyright: ignore[reportAttributeAccessIssue]
        purchased_id=id_from_href(purchased_at) if purchased_at else None,  # pyright: ignore[reportArgumentType]
        comment=comment,
        serving=serving.text if serving else None,  # pyright: ignore[reportAttributeAccessIssue]
        user_rating=data_rating,
        friends=friends,
    )


def extract_checkin_elements(
    element: Element,
) -> tuple[Element, Element, Element, Element | None]:
    """Extract four linked elements in a checkin.

    Args:
        element (Element): checkin element

    Raises:
        ValueError: element passed didn't contain 3-4 <a> tags

    Returns:
        tuple[Element, Element, Element, Element]: user, beer, brewery, location
    """
    elements = element.find(".top .text a")

    if len(elements) == 3:  # noqa: PLR2004  # pyright: ignore[reportArgumentType]
        return [*elements, None]  # pyright: ignore[reportGeneralTypeIssues, reportReturnType]
    if len(elements) == 4:  # noqa: PLR2004  # pyright: ignore[reportArgumentType]
        return elements  # pyright: ignore[reportReturnType]

    msg = f"Wanted 3 or 4 <a> elements (not {len(elements)}) in {element.html}"  # pyright: ignore[reportArgumentType]
    raise ValueError(msg)



"""Test brewery scraping."""

from __future__ import annotations

import pytest
from untappd_scraper.brewery import Brewery
from utpd_models_web.constants import UNTAPPD_BREWERY_HISTORY_SIZE

# ----- Tests -----


@pytest.mark.webtest
def test_brewery_lookup() -> None:
    result = Brewery.from_name("big niles")

    assert result
    assert "Niles" in result.name
    assert "Niles" in result.brewery_name
    assert "dalmeny" in result.address.casefold()


@pytest.mark.webtest
def test_brewery_from_slug() -> None:
    result = Brewery.from_url("BigNilesBrewingCo")

    assert result
    assert "Niles" in result.name
    assert "Niles" in result.brewery_name
    assert "dalmeny" in result.address.casefold()
    assert result.brewery_id == 433022


@pytest.mark.webtest
def test_brewery_buckettys() -> None:
    result = Brewery(484738)  # Bucketty's

    assert result
    assert "bucketty" in result.name.casefold()
    assert "brookvale" in result.address.casefold()
    assert result.style
    assert result.description


@pytest.mark.webtest
def test_brewery_invalid() -> None:
    with pytest.raises(ValueError, match="Invalid brewery ID"):
        Brewery(1234567890)


@pytest.mark.webtest
def test_brewery_invalid_name() -> None:
    result = Brewery.from_name("Koneko and Sylvester")

    assert result is None


@pytest.mark.webtest
def test_brewery_invalid_slug() -> None:
    result = Brewery.from_url("Koneko-Sylvester")

    assert result is None


@pytest.mark.webtest
def test_brewery_details() -> None:
    result = Brewery(446724).details  # MCBC

    assert result
    assert "mountain" in result.name.casefold()
    assert result.description
    assert result.brewery_url
    assert result.brewery_id == 446724


@pytest.mark.webtest
def test_brewery_activity() -> None:
    result = Brewery(446724).activity  # MCBC

    assert result
    assert len(result) == UNTAPPD_BREWERY_HISTORY_SIZE
    assert all("mountain" in r.brewery_name.casefold() for r in result)
    assert any(r.friends for r in result)
    assert any(r.purchased_at for r in result)
    assert all("purchased" not in r.purchased_at.casefold() for r in result if r.purchased_at)
    assert all(r.beer_label_url for r in result)



"""Utilities to process web data."""

from __future__ import annotations

from typing import TYPE_CHECKING

import parse
from bs4 import BeautifulSoup, Tag
from dateutil.parser import parse as parse_date
from httpx import URL
from utpd_models_web.constants import UNTAPPD_BASE_URL

from untappd_scraper.client import client
from untappd_scraper.logging_config import logger

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Mapping
    from datetime import datetime

    from requests_html import Element

logger.info("Loading web...")


def url_of(  # noqa: PLR0913
    user_id: str = "",
    *,
    venue_id: str = "",
    brewery_id: str | int = "",
    beer_id: str | int = "",
    search: bool = False,
    page: str = "",
    query: Mapping[str, str | int] | None = None,
) -> str:
    """Return the URL for a user's main page.

    Args:
        user_id (str): user ID
        venue_id (str): venue ID
        brewery_id (str): brewery ID
        beer_id (str|int): beer ID
        search (bool): whether to search
        page (str): specific page
        query (dict[str, str]|None): filter for page

    Returns:
        str: url to load to get user's main page
    """
    base = UNTAPPD_BASE_URL

    if user_id:
        base += f"user/{user_id}"
    elif venue_id:
        base += f"venue/{venue_id}"
    elif brewery_id:
        base += f"brewery/{brewery_id}"
    elif beer_id:
        base += f"beer/{beer_id}"
    elif search:
        base += "search"
    if page:
        if not base.endswith("/"):
            base += "/"
        base += f"{page.strip('/')}"

    base = base.strip("/")

    url = URL(base)

    if query:
        url = url.copy_merge_params(query)

    return str(url)


def make_soup(url: str) -> tuple[BeautifulSoup, bool]:
    """Fetch a page and return the soup.

    Args:
        url (str): URL to fetch

    Returns:
        tuple[BeautifulSoup, bool]: soup object and whether it was from cache
    """
    resp = client.get(url, timeout=10)
    resp.raise_for_status()

    logger.debug("Making soup from {}", url)

    return BeautifulSoup(resp.text, "html.parser"), resp.extensions.get("from_cache", False)


def id_from_href(element: Element) -> int:
    """Extract last past of a URL, which is an ID.

    Args:
        element (Element): element to extract href from

    Raises:
        ValueError: last part of url wasn't an integer

    Returns:
        int: last part of href, which is an id
    """
    last_bit: str = slug_from_href(element)
    try:
        return int(last_bit)
    except ValueError as excp:  # pragma: no cover
        msg = f"Cannot extract integer id from {last_bit=} element ({element.html})"
        raise ValueError(msg) from excp


def slug_from_href(element: Element) -> str:
    """Extract last past of a URL, which is a slug.

    Args:
        element (Element): element to extract href from

    Returns:
        str: last part of href
    """
    href: str = element.attrs["href"].removesuffix("/")
    return href.rpartition("/")[-1]


def end_of_href(tag: Tag) -> str | None:
    """Extract last past of a URL, which is a slug.

    Args:
        tag (Tag): html element containing an href

    Returns:
        str: last part of href
    """
    href = (tag.a and tag.a.get("href")) or tag.get("href")

    if not href or "/" not in href:
        logger.debug("No href found in element: {}", tag)
        return None

    href = str(href).removesuffix("/")

    return href.rpartition("/")[-1]


def date_from_details(details: Tag) -> datetime | None:
    """Extract a date that may be present.

    Args:
        details (Tag): html element containing date

    Returns:
        datetime: parsed date
    """
    dt = details.select_one(".date-time")

    return parse_date(dt.text) if dt else None


def date_from_data_href(
    element: Element, label: str, *, date_only: bool = False
) -> datetime | None:
    """Extract a date that may be present.

    Args:
        element (Element): html element containing date
        label (str): data-href value
        date_only (bool): only parse the date (not the time)?

    Returns:
        datetime: parsed date
    """
    if href := element.find(f'.date [data-href="{label}"]', first=True):
        dt = parse_date(href.text).astimezone()  # pyright: ignore[reportAttributeAccessIssue]
        return dt.date() if date_only else dt  # pyright: ignore[reportReturnType]
    return None


def parse_abv(abv_data: str) -> float | None:
    """Parse the ABV data buried in text.

    Args:
        abv_data (str): text version, eg '7.5% ABV'

    Returns:
        float: ABV as a number
    """
    parsed_abv = parse.search("{:g}%", abv_data)
    if not isinstance(parsed_abv, parse.Result):
        return None  # pragma: no cover

    try:
        return float(parsed_abv[0])
    except TypeError:  # pragma: no cover
        return None


def parse_ibu(ibu_data: str) -> int | None:
    """Parse the IBU data buried in text.

    Args:
        ibu_data (str): text version, eg '15 IBU'

    Returns:
        int: IBU as a number
    """
    parsed_ibu = parse.search("{:d} IBU", ibu_data)
    if not isinstance(parsed_ibu, parse.Result):
        return None  # pragma: no cover

    try:
        return int(parsed_ibu[0])
    except TypeError:  # pragma: no cover
        return None


def parsed_value(fmt: str, string: str) -> int | float | None:
    """Search for a value in a string and return it.

    Args:
        fmt (str): parse module's format string
        string (str): string to search for fmt pattern

    Returns:
        found value, if any
    """
    match = parse.search(fmt, string)
    return match[0] if isinstance(match, parse.Result) else None


def parse_rating(which: str, ratings: str) -> float | None:
    """Find the requested rating in the supplied string.

    Args:
        which (str): which rating to look for
        ratings (str): full rating string with all ratings

    Returns:
        float: rating found, if any
    """
    ratings_match = parse.search(which + " rating ({:g})", ratings)
    return ratings_match[0] if ratings_match else None  # pyright: ignore[reportIndexIssue]


def get_text(soup: BeautifulSoup | Tag, selector: str) -> str:
    """Get text from a soup element.

    Args:
        soup (BeautifulSoup): soup to search
        selector (str): CSS selector to find

    Returns:
        str: text found
    """
    el = soup.select_one(selector)
    return el.text.strip() if el else ""


def extract_checkin_elements_bs4(tag: Tag) -> tuple[Tag, Tag, Tag, Tag | None]:
    """Extract four linked elements in a checkin.

    Args:
        tag (Tag): checkin element

    Raises:
        ValueError: element passed didn't contain 3-4 <a> tags

    Returns:
        tuple[Tag, Tag, Tag, Tag]: user, beer, brewery, location
    """
    elements = tag.select(".top .text a")

    if len(elements) == 3:  # noqa: PLR2004
        return (elements[0], elements[1], elements[2], None)
    if len(elements) == 4:  # noqa: PLR2004
        return (elements[0], elements[1], elements[2], elements[3])

    msg = f"Wanted 3 or 4 <a> elements (not {len(elements)}) in {tag}"
    raise ValueError(msg)
